# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`HBA` (Host Bus Adapter) is a logical entity that provides a
:term:`Partition` with access to external storage area networks (SANs) through
an :term:`FCP Adapter`. More specifically, an HBA connects a Partition with an
:term:`Adapter Port` on an FCP Adapter.

HBA resources are contained in Partition resources.

HBA resources only exist in :term:`CPCs <CPC>` that are in DPM mode and when
the "dpm-storage-management" :ref:`firmware feature <firmware features>` is not
enabled. See section :ref:`Storage Groups` for details. When the
"dpm-storage-management" firmware feature is enabled, :term:`virtual HBAs <HBA>`
are represented as :term:`Virtual Storage Resource` resources.
"""


import copy

from ._manager import BaseManager
from ._resource import BaseResource
from ._logging import logged_api_call
from ._utils import RC_HBA

__all__ = ['HbaManager', 'Hba']


class HbaManager(BaseManager):
    """
    Manager providing access to the :term:`HBAs <HBA>` in a particular
    :term:`Partition`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable of a
    :class:`~zhmcclient.Partition` object (in DPM mode):

    * :attr:`~zhmcclient.Partition.hbas`

    HMC/SE version requirements:

    * SE version >= 2.13.1 without
      :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, partition):
        # This function should not go into the docs.
        # Parameters:
        #   partition (:class:`~zhmcclient.Partition`):
        #     Partition defining the scope for this manager.

        super().__init__(
            resource_class=Hba,
            class_name=RC_HBA,
            session=partition.manager.session,
            parent=partition,
            base_uri=f'{partition.uri}/hbas',
            oid_prop='element-id',
            uri_prop='element-uri',
            name_prop='name',
            query_props=[],
            list_has_name=False)

    @property
    def partition(self):
        """
        :class:`~zhmcclient.Partition`: :term:`Partition` defining the scope
        for this manager.
        """
        return self._parent

    @logged_api_call
    def list(self, full_properties=False, filter_args=None):
        """
        List the HBAs in this Partition.

        Any resource property may be specified in a filter argument. For
        details about filter arguments, see :ref:`Filtering`.

        The listing of resources is handled in an optimized way:

        * If this manager is enabled for :ref:`auto-updating`, a locally
          maintained resource list is used (which is automatically updated via
          inventory notifications from the HMC) and the provided filter
          arguments are applied.

        * Otherwise, if the filter arguments specify the resource name as a
          single filter argument with a straight match string (i.e. without
          regular expressions), an optimized lookup is performed based on a
          locally maintained name-URI cache.

        * Otherwise, the corresponding array property for this resource in the
          parent object is used to list the resources, and the provided filter
          arguments are applied.

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this Partition.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

        Returns:

          : A list of :class:`~zhmcclient.Hba` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        return self._list_with_parent_array(
            self.partition, 'hba-uris', full_properties, filter_args)

    @logged_api_call
    def create(self, properties):
        """
        Create and configure an HBA in this Partition.

        The HBA must be backed by an adapter port on an FCP adapter.

        The backing adapter port is specified in the "properties" parameter of
        this method by setting the "adapter-port-uri" property to the URI of
        the backing adapter port.

        The value for the "adapter-port-uri" property can be determined from a
        given adapter name and port index as shown in the following example
        code (omitting any error handling):

        .. code-block:: python

            partition = ...  # Partition object for the new HBA

            adapter_name = 'FCP #1'  # name of adapter with backing port
            adapter_port_index = 0   # port index of backing port

            adapter = partition.manager.cpc.adapters.find(name=adapter_name)

            port = adapter.ports.find(index=adapter_port_index)

            properties['adapter-port-uri'] = port.uri

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to this Partition.
        * Object-access permission to the backing Adapter for the new HBA.
        * Task permission to the "Partition Details" task.

        Parameters:

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Create HBA' in the :term:`HMC API` book.

        Returns:

          Hba:
            The resource object for the new HBA.
            The object will have its 'element-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(self.partition.uri + '/hbas',
                                   body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        uri = props[self._uri_prop]
        hba = Hba(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return hba


class Hba(BaseResource):
    """
    Representation of an :term:`HBA`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    For the properties of an HBA resource, see section
    'Data model - HBA Element Object' in section 'Partition object' in the
    :term:`HMC API` book.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.HbaManager`).

    HMC/SE version requirements:

    * SE version >= 2.13.1 without
      :ref:`firmware feature <firmware features>` "dpm-storage-management"
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        # Parameters:
        #   manager (:class:`~zhmcclient.HbaManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, HbaManager), (
            f"Hba init: Expected manager type {HbaManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)

    @logged_api_call
    def delete(self):
        """
        Delete this HBA.

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the Partition containing this HBA.
        * Task permission to the "Partition Details" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self._uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))

        parent_hba_uris = self.manager.parent.get_properties_local(
            'hba-uris')
        if parent_hba_uris:
            try:
                parent_hba_uris.remove(self._uri)
            except ValueError:
                pass

        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this HBA.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the Partition containing this HBA.
        * **TBD: Verify:** Object-access permission to the backing Adapter for
          this HBA.
        * Task permission to the "Partition Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model - HBA Element Object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def reassign_port(self, port):
        """
        Reassign this HBA to a new underlying :term:`FCP port`.

        This method performs the HMC operation "Reassign Storage Adapter Port".

        HMC/SE version requirements:

        * SE version >= 2.13.1 without
          :ref:`firmware feature <firmware features>` "dpm-storage-management"

        Authorization requirements:

        * Object-access permission to the Partition containing this HBA.
        * Object-access permission to the Adapter with the new Port.
        * Task permission to the "Partition Details" task.

        Parameters:

          port (:class:`~zhmcclient.Port`): :term:`FCP port` to be used.

        Raises:

          :exc:`~zhmcclient.HTTPError`: See the HTTP status and reason codes of
            operation "Reassign Storage Adapter Port" in the :term:`HMC API`
            book.
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        body = {'adapter-port-uri': port.uri}
        self.manager.session.post(
            self._uri + '/operations/reassign-storage-adapter-port',
            resource=self, body=body)
        self.update_properties_local(body)



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that lists the Reset/Image/Load Activation Profiles on a CPC in classic
mode.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in classic mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': False})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in classic mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    # Reset activation profiles
    print(f"Listing reset activation profiles for CPC {cpc.name} ...")
    try:
        profiles = cpc.reset_activation_profiles.list()
    except zhmcclient.Error as exc:
        print("Error: Cannot list reset activation profiles for CPC "
              f"{cpc.name}: {exc.__class__.__name__}: {exc}")
        sys.exit(1)
    for profile in profiles:
        print(profile.name, profile.get_property('element-uri'))

    # Image activation profiles
    print(f"Listing image activation profiles for CPC {cpc.name} ...")
    try:
        profiles = cpc.image_activation_profiles.list()
    except zhmcclient.Error as exc:
        print("Error: Cannot list image activation profiles for CPC "
              f"{cpc.name}: {exc.__class__.__name__}: {exc}")
        sys.exit(1)
    for profile in profiles:
        print(profile.name, profile.get_property('element-uri'))

    # Load activation profiles
    print(f"Listing load activation profiles for CPC {cpc.name} ...")
    try:
        profiles = cpc.load_activation_profiles.list()
    except zhmcclient.Error as exc:
        print("Error: Cannot list load activation profiles for CPC "
              f"{cpc.name}: {exc.__class__.__name__}: {exc}")
        sys.exit(1)
    for profile in profiles:
        print(profile.name, profile.get_property('element-uri'))

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Unit tests for _timestats module.
"""


import time
import pytest

from zhmcclient import TimeStatsKeeper, TimeStats


PRINT_HEADER = \
    "Time statistics (times in seconds):\n" \
    "Count  Average  Minimum  Maximum  Operation name"

PRINT_HEADER_DISABLED = \
    "Time statistics (times in seconds):\n" \
    "Disabled."


def time_abs_delta(t1, t2):
    """
    Return the positive difference between two float values.
    """
    return abs(t1 - t2)


def measure(stats, duration):
    """
    Return the measured duration of one begin/end cycle of a TimeStats object
    of a given intended duration.  The runtime behavior may be that the actual
    duration is larger than the intended duration, therefore we need to measure
    the actual duration.
    """
    begin = time.time()
    stats.begin()
    time.sleep(duration)
    stats.end()
    end = time.time()
    return end - begin


def test_timestatskeeper_enabling():
    """Test enabling and disabling."""

    keeper = TimeStatsKeeper()

    assert not keeper.enabled, \
        "Verify that initial state is disabled"

    keeper.disable()
    assert not keeper.enabled, \
        "Verify that disabling a disabled keeper works"

    keeper.enable()
    assert keeper.enabled, \
        "Verify that enabling a disabled keeper works"

    keeper.enable()
    assert keeper.enabled, \
        "Verify that enabling an enabled keeper works"

    keeper.disable()
    assert not keeper.enabled, \
        "Verify that disabling an enabled keeper works"


def test_timestatskeeper_get():
    """Test getting time statistics."""

    keeper = TimeStatsKeeper()
    snapshot_length = len(keeper.snapshot())
    assert snapshot_length == 0, \
        "Verify that initial state has no time statistics. " \
        f"Actual number = {snapshot_length}"

    stats = keeper.get_stats('foo')
    snapshot_length = len(keeper.snapshot())
    assert snapshot_length == 0, \
        "Verify that getting a new stats with a disabled keeper results " \
        f"in no time statistics. Actual number = {snapshot_length}"
    assert stats.keeper == keeper
    assert stats.name == "disabled"  # stats for disabled keeper
    assert stats.count == 0
    assert stats.avg_time == 0
    assert stats.min_time == float('inf')
    assert stats.max_time == 0

    keeper.enable()

    stats = keeper.get_stats('foo')
    snapshot_length = len(keeper.snapshot())
    assert snapshot_length == 1, \
        "Verify that getting a new stats with an enabled keeper results " \
        f"in one time statistics. Actual number = {snapshot_length}"

    assert stats.keeper == keeper
    assert stats.name == 'foo'
    assert stats.count == 0
    assert stats.avg_time == 0
    assert stats.min_time == float('inf')
    assert stats.max_time == 0

    keeper.get_stats('foo')
    snapshot_length = len(keeper.snapshot())
    assert snapshot_length == 1, \
        "Verify that getting an existing stats with an enabled keeper " \
        "results in the same number of time statistics. " \
        f"Actual number = {snapshot_length}"


def test_timestatskeeper_measure_enabled():
    """Test measuring time with enabled keeper."""

    keeper = TimeStatsKeeper()
    keeper.enable()

    # TimeStatsKeeper on Windows has only a precision of 1/60 sec
    duration = 1.6
    delta = duration / 10.0

    stats = keeper.get_stats('foo')
    dur = measure(stats, duration)

    stats_dict = keeper.snapshot()
    for op_name in stats_dict:
        stats = stats_dict[op_name]
        assert stats.count == 1
        assert time_abs_delta(stats.avg_time, dur) < delta, (
            f"avg time: actual: {stats.avg_time}, expected: {dur}, "
            f"delta: {delta}")
        assert time_abs_delta(stats.min_time, dur) < delta, (
            f"min time: actual: {stats.min_time}, expected: {dur}, "
            f"delta: {delta}")
        assert time_abs_delta(stats.max_time, dur) < delta, (
            f"max time: actual: {stats.max_time}, expected: {dur}, "
            f"delta: {delta}")

    stats.reset()
    assert stats.count == 0
    assert stats.avg_time == 0
    assert stats.min_time == float('inf')
    assert stats.max_time == 0


def test_timestatskeeper_measure_disabled():
    """Test measuring time with disabled keeper."""

    keeper = TimeStatsKeeper()

    duration = 0.2

    stats = keeper.get_stats('foo')
    assert stats.name == 'disabled'

    stats.begin()
    time.sleep(duration)
    stats.end()

    stats_dict = keeper.snapshot()
    for op_name in stats_dict:
        stats = stats_dict[op_name]
        assert stats.count == 0
        assert stats.avg_time == 0
        assert stats.min_time == float('inf')
        assert stats.max_time == 0


def test_timestatskeeper_snapshot():
    """Test that snapshot() takes a stable snapshot."""

    keeper = TimeStatsKeeper()
    keeper.enable()

    duration = 0.2

    stats = keeper.get_stats('foo')

    # produce a first data item
    stats.begin()
    time.sleep(duration)
    stats.end()

    # take the snapshot
    snap_stats_dict = keeper.snapshot()

    # produce a second data item
    stats.begin()
    time.sleep(duration)
    stats.end()

    # verify that only the first data item is in the snapshot
    for op_name in snap_stats_dict:
        snap_stats = snap_stats_dict[op_name]
        assert snap_stats.count == 1

    # verify that both data items are in the original stats object
    assert stats.count == 2


def test_timestatskeeper_measure_avg_min_max():
    """Test measuring avg min max values."""

    keeper = TimeStatsKeeper()
    keeper.enable()

    # TimeStatsKeeper on Windows has only a precision of 1/60 sec
    durations = (0.6, 1.2, 1.5)
    delta = 0.08
    count = len(durations)

    stats = keeper.get_stats('foo')
    m_durations = []
    for duration in durations:
        m_durations.append(measure(stats, duration))

    min_dur = min(m_durations)
    max_dur = max(m_durations)
    avg_dur = sum(m_durations) / float(count)

    stats_dict = keeper.snapshot()
    for op_name in stats_dict:
        stats = stats_dict[op_name]
        assert stats.count == 3
        assert time_abs_delta(stats.avg_time, avg_dur) < delta, (
            f"avg time: actual: {stats.avg_time}, expected: {avg_dur}, "
            f"delta: {delta}")
        assert time_abs_delta(stats.min_time, min_dur) < delta, (
            f"min time: actual: {stats.min_time}, expected: {avg_dur}, "
            f"delta: {delta}")
        assert time_abs_delta(stats.max_time, max_dur) < delta, (
            f"max time: actual: {stats.max_time}, expected: {avg_dur}, "
            f"delta: {delta}")


def test_timestatskeeper_only_end():
    """Test that invoking end() before begin() has ever been called raises
    a RuntimeError exception."""

    keeper = TimeStatsKeeper()
    keeper.enable()
    stats = keeper.get_stats('foo')

    with pytest.raises(RuntimeError):
        stats.end()


def test_timestatskeeper_end_after_end():
    """Test that invoking end() after a begin/end sequence raises
    a RuntimeError exception."""

    keeper = TimeStatsKeeper()
    keeper.enable()
    stats = keeper.get_stats('foo')

    stats.begin()
    time.sleep(0.01)
    stats.end()

    with pytest.raises(RuntimeError):
        stats.end()


def test_timestatskeeper_str_empty():
    """Test TimestatsKeeper.__str__() for an empty enabled keeper."""

    keeper = TimeStatsKeeper()
    keeper.enable()
    s = str(keeper)
    assert s == PRINT_HEADER


def test_timestatskeeper_str_disabled():
    """Test TimestatsKeeper.__str__() for a disabled keeper."""

    keeper = TimeStatsKeeper()
    s = str(keeper)
    assert s == PRINT_HEADER_DISABLED


def test_timestatskeeper_str_one():
    """Test TimestatsKeeper.__str__() for an enabled keeper with one data
    item."""

    keeper = TimeStatsKeeper()
    keeper.enable()

    duration = 0.1

    stats = keeper.get_stats('foo')

    # produce a data item
    stats.begin()
    time.sleep(duration)
    stats.end()

    s = str(keeper)
    assert s.startswith(PRINT_HEADER), \
        f"Unexpected str(keeper): {s!r}"
    num_lines = len(s.split('\n'))
    assert num_lines == 3, \
        f"Unexpected str(keeper): {s!r}"


def test_timestats_str():
    """Test Timestats.__str__()."""

    keeper = TimeStatsKeeper()
    timestats = TimeStats(keeper, "foo")

    s = str(timestats)
    assert s.startswith("TimeStats:"), \
        f"Unexpected str(timestats): {s!r}"
    num_lines = len(s.split('\n'))
    assert num_lines == 1, \
        f"Unexpected str(timestats): {s!r}"



# Copyright 2016,2021 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Utility functions.
"""


import re
from collections import OrderedDict
from collections.abc import Mapping, MutableSequence, Iterable
from datetime import datetime
import warnings

from dateutil import parser
import pytz
from requests.utils import quote

from ._exceptions import HTTPError

__all__ = ['datetime_from_timestamp', 'timestamp_from_datetime']


_EPOCH_DT = datetime(1970, 1, 1, 0, 0, 0, 0, pytz.utc)

# Resource class names.
# These are the values of the 'class' property on the resource objects.
#
# For CPCs in DPM mode:
RC_ADAPTER = 'adapter'
RC_CAPACITY_GROUP = 'capacity-group'
RC_HBA = 'hba'
RC_NIC = 'nic'
RC_PARTITION = 'partition'
RC_NETWORK_PORT = 'network-port'
RC_STORAGE_PORT = 'storage-port'
RC_STORAGE_TEMPLATE = 'storage-template'
RC_STORAGE_GROUP = 'storage-group'
RC_STORAGE_TEMPLATE_VOLUME = 'storage-template-volume'
RC_STORAGE_VOLUME = 'storage-volume'
RC_VIRTUAL_FUNCTION = 'virtual-function'
RC_VIRTUAL_STORAGE_RESOURCE = 'virtual-storage-resource'
RC_VIRTUAL_SWITCH = 'virtual-switch'
RC_STORAGE_SITE = 'storage-site'
RC_STORAGE_FABRIC = 'storage-fabric'
RC_STORAGE_SWITCH = 'storage-switch'
RC_STORAGE_SUBSYSTEM = 'storage-subsystem'
RC_STORAGE_PATH = 'storage-path'
RC_STORAGE_CONTROL_UNIT = 'storage-control-unit'
RC_VIRTUAL_TAPE_RESOURCE = 'virtual-tape-resource'
RC_TAPE_LINK = 'tape-link'
RC_TAPE_LIBRARY = 'tape-library'
RC_CERTIFICATE = 'certificate'
#
# For CPCs in classic mode:
RC_RESET_ACTIVATION_PROFILE = 'reset-activation-profile'
RC_IMAGE_ACTIVATION_PROFILE = 'image-activation-profile'
RC_LOAD_ACTIVATION_PROFILE = 'load-activation-profile'
RC_LDAP_SERVER_DEFINITION = 'ldap-server-definition'
RC_LOGICAL_PARTITION = 'logical-partition'
#
# For CPCs in any  mode and resources independent of CPCs:
RC_CONSOLE = 'console'
RC_CPC = 'cpc'
RC_PASSWORD_RULE = 'password-rule'  # nosec B105
RC_TASK = 'task'
RC_USER_PATTERN = 'user-pattern'
RC_USER_ROLE = 'user-role'
RC_USER = 'user'
RC_GROUP = 'group'

# Resource classes that are children of zhmcclient.Cpc
RC_CHILDREN_CPC = (
    RC_PARTITION,
    RC_ADAPTER,
    RC_VIRTUAL_SWITCH,
    RC_CAPACITY_GROUP,
    RC_LOGICAL_PARTITION,
    RC_RESET_ACTIVATION_PROFILE,
    RC_IMAGE_ACTIVATION_PROFILE,
    RC_LOAD_ACTIVATION_PROFILE,
)
# Resource classes that are children of zhmcclient.Console
RC_CHILDREN_CONSOLE = (
    RC_STORAGE_GROUP,
    RC_STORAGE_TEMPLATE,
    RC_PASSWORD_RULE,
    RC_TASK,
    RC_USER_PATTERN,
    RC_USER_ROLE,
    RC_USER,
    RC_LDAP_SERVER_DEFINITION,
    RC_CPC,  # For unmanaged CPCs
)
# Resource classes that are children of zhmcclient.Client (= top level)
RC_CHILDREN_CLIENT = (
    RC_CPC,  # For managed CPCs
    RC_CONSOLE,
)


# Valid resource class names
#:
VALID_RESOURCE_CLASSES = frozenset([
    RC_ADAPTER,
    RC_CAPACITY_GROUP,
    RC_HBA,
    RC_NIC,
    RC_PARTITION,
    RC_NETWORK_PORT,
    RC_STORAGE_PORT,
    RC_STORAGE_TEMPLATE,
    RC_STORAGE_GROUP,
    RC_STORAGE_TEMPLATE_VOLUME,
    RC_STORAGE_VOLUME,
    RC_VIRTUAL_FUNCTION,
    RC_VIRTUAL_STORAGE_RESOURCE,
    RC_VIRTUAL_SWITCH,
    RC_STORAGE_SITE,
    RC_STORAGE_FABRIC,
    RC_STORAGE_SWITCH,
    RC_STORAGE_SUBSYSTEM,
    RC_STORAGE_PATH,
    RC_STORAGE_CONTROL_UNIT,
    RC_VIRTUAL_TAPE_RESOURCE,
    RC_TAPE_LINK,
    RC_TAPE_LIBRARY,
    RC_RESET_ACTIVATION_PROFILE,
    RC_IMAGE_ACTIVATION_PROFILE,
    RC_LOAD_ACTIVATION_PROFILE,
    RC_LDAP_SERVER_DEFINITION,
    RC_LOGICAL_PARTITION,
    RC_CONSOLE,
    RC_CPC,
    RC_PASSWORD_RULE,
    RC_TASK,
    RC_USER_PATTERN,
    RC_USER_ROLE,
    RC_USER,
    RC_GROUP,
])


def _indent(text, amount, ch=' '):
    """Return the indent text, where each line is indented by `amount`
    characters `ch`."""
    padding = amount * ch
    return ''.join(padding + line for line in text.splitlines(True))


def repr_text(text, indent):
    """Return a debug representation of a multi-line text (e.g. the result
    of another repr...() function)."""
    if text is None:
        return 'None'
    ret = _indent(text, amount=indent)
    return ret.lstrip(' ')


def repr_list(lst, indent):
    """Return a debug representation of a list or tuple."""
    # pprint represents lists and tuples in one row if possible. We want one
    # per row, so we iterate ourselves.
    if lst is None:
        return 'None'
    if isinstance(lst, MutableSequence):
        bm = '['
        em = ']'
    elif isinstance(lst, Iterable):
        bm = '('
        em = ')'
    else:
        raise TypeError(f"Object must be an iterable, but is a {type(lst)}")
    ret = bm + '\n'
    for value in lst:
        ret += _indent(f'{value!r},\n', 2)
    ret += em
    ret = repr_text(ret, indent=indent)
    return ret.lstrip(' ')


def repr_dict(dct, indent):
    """Return a debug representation of a dict or OrderedDict."""
    # pprint represents OrderedDict objects using the tuple init syntax,
    # which is not very readable. Therefore, dictionaries are iterated over.
    if dct is None:
        return 'None'
    if not isinstance(dct, Mapping):
        raise TypeError(f"Object must be a mapping, but is a {type(dct)}")
    if isinstance(dct, OrderedDict):
        kind = 'ordered'
        ret = '%s {\n' % kind  # non standard syntax for the kind indicator
        for key in dct.keys():
            value = dct[key]
            ret += _indent(f'{key!r}: {value!r},\n', 2)
    else:  # dict
        kind = 'sorted'
        ret = '%s {\n' % kind  # non standard syntax for the kind indicator
        for key in sorted(dct.keys()):
            value = dct[key]
            ret += _indent(f'{key!r}: {value!r},\n', 2)
    ret += '}'
    ret = repr_text(ret, indent=indent)
    return ret.lstrip(' ')


def repr_timestamp(timestamp):
    """Return a debug representation of an HMC timestamp number."""
    if timestamp is None:
        return 'None'
    dt = datetime_from_timestamp(timestamp)
    ret = f"{timestamp} ({dt.strftime('%Y-%m-%d %H:%M:%S.%f %Z')})"
    return ret


def repr_manager(manager, indent):
    """Return a debug representation of a manager object."""
    return repr_text(repr(manager), indent=indent)


def repr_obj_id(obj):
    """
    Return a string that shows the class and ID value of the input object,
    for use in repr() functions.
    """
    return f"{obj.__class__.__name__} at 0x{id(obj):08x}"


def datetime_from_timestamp(ts, tzinfo=pytz.utc):
    """
    Convert an :term:`HMC timestamp number <timestamp>` into a
    :class:`~py:datetime.datetime` object. The resulting object will be
    timezone-aware and will be represented in the specified timezone,
    defaulting to UTC.

    The HMC timestamp number must be non-negative. This means the special
    timestamp value -1 cannot be represented as datetime and will cause
    ``ValueError`` to be raised.

    The date and time range supported by this function has the following
    bounds:

    * The upper bounds is determined by :attr:`py:datetime.datetime.max` and
      additional limitations, as follows:

      * 9999-12-31 23:59:59 UTC, for 32-bit and 64-bit CPython on Linux and
        OS-X.
      * 3001-01-01 07:59:59 UTC, for 32-bit and 64-bit CPython on Windows, due
        to a limitation in `gmtime() in Visual C++
        <https://docs.microsoft.com/en-us/cpp/c-runtime-library/reference/gmtime-gmtime32-gmtime64>`_.
      * 2038-01-19 03:14:07 UTC, for some 32-bit Python implementations,
        due to the `Year 2038 problem
        <https://en.wikipedia.org/wiki/Year_2038_problem>`_.

    * The lower bounds is the UNIX epoch: 1970-01-01 00:00:00 UTC.

    Parameters:

      ts (:term:`timestamp`):
        Point in time as an HMC timestamp number.

        Must not be `None`.

      tzinfo (:class:`py:datetime.tzinfo`):
        Timezone in which the returned object will be represented.
        This may be any object derived from :class:`py:datetime.tzinfo`,
        including but not limited to objects returned by
        :func:`pytz.timezone`.

        Note that this parameter does not affect how the HMC timestamp value is
        interpreted; i.e. the effective point in time represented by the
        returned object is not affected. What is affected by this parameter
        is for example the timezone in which the point in time is shown when
        printing the returned object.

        Must not be `None`.

    Returns:

      :class:`~py:datetime.datetime`:
        Point in time as a timezone-aware Python datetime object for the
        specified timezone.

    Raises:

        ValueError
    """
    # Note that in Python 2, "None < 0" is allowed and will return True,
    # therefore we do an extra check for None.
    if ts is None:
        raise ValueError("HMC timestamp value must not be None.")
    if ts < 0:
        raise ValueError(
            f"Negative HMC timestamp value {ts} cannot be represented as "
            "datetime.")
    epoch_seconds = ts // 1000
    delta_microseconds = ts % 1000 * 1000
    if tzinfo is None:
        raise ValueError("Timezone must not be None.")
    try:
        dt = datetime.fromtimestamp(epoch_seconds, tzinfo)
    except (ValueError, OSError) as exc:
        new_exc = ValueError(str(exc))
        new_exc.__cause__ = None
        raise new_exc  # ValueError
    dt = dt.replace(microsecond=delta_microseconds)
    return dt


def timestamp_from_datetime(dt):
    """
    Convert a :class:`~py:datetime.datetime` object into an
    :term:`HMC timestamp number <timestamp>`.

    The date and time range supported by this function has the following
    bounds:

    * The upper bounds is :attr:`py:datetime.datetime.max`, as follows:

      * 9999-12-31 23:59:59 UTC, for 32-bit and 64-bit CPython on Linux and
        OS-X.
      * 2038-01-19 03:14:07 UTC, for some 32-bit Python implementations,
        due to the `Year 2038 problem
        <https://en.wikipedia.org/wiki/Year_2038_problem>`_.

    * The lower bounds is the UNIX epoch: 1970-01-01 00:00:00 UTC.

    Parameters:

      dt (:class:`~py:datetime.datetime`):
        Point in time as a Python datetime object. The datetime object may be
        timezone-aware or timezone-naive. If timezone-naive, the UTC timezone
        is assumed.

        Must not be `None`.

    Returns:

      :term:`timestamp`:
        Point in time as an HMC timestamp number.

    Raises:

        ValueError
    """
    if dt is None:
        raise ValueError("datetime value must not be None.")
    if dt.tzinfo is None:
        # Apply default timezone to the timezone-naive input
        dt = pytz.utc.localize(dt)
    epoch_seconds = (dt - _EPOCH_DT).total_seconds()
    ts = int(epoch_seconds * 1000)
    return ts


def append_query_parms(query_parms, prop_name, prop_match):
    """
    Append prop_name=prop_match to query_parms. prop_match may be a list.
    """
    if isinstance(prop_match, (list, tuple)):
        for pm in prop_match:
            append_query_parms(query_parms, prop_name, pm)
    else:
        # Just in case, we also escape the property name
        parm_name = quote(prop_name, safe='')
        parm_value = quote(str(prop_match), safe='')
        qp = f'{parm_name}={parm_value}'
        query_parms.append(qp)


def divide_filter_args(query_props, filter_args):
    """
    Divide the filter arguments into filter query parameters for filtering
    on the server side, and the remaining client-side filters.

    Parameters:

      query_props (iterable of strings):
        Names of resource properties that are supported as filter query
        parameters in the WS API operation (i.e. server-side filtering).

        May be `None`.

        If the support for a resource property changes within the set of
        HMC versions that support this type of resource, this list must
        represent the version of the HMC this session is connected to.

      filter_args (dict):
        Filter arguments that narrow the list of returned resources to
        those that match the specified filter arguments. For details, see
        :ref:`Filtering`.

        `None` causes no filtering to happen, i.e. all resources are
        returned.

    Returns:

      : tuple (query_parms, client_filter_args)
    """
    query_parms = []  # query parameter strings
    client_filter_args = {}

    if filter_args is not None:
        for prop_name in filter_args:
            prop_match = filter_args[prop_name]
            if prop_name in query_props:
                append_query_parms(query_parms, prop_name, prop_match)
            else:
                client_filter_args[prop_name] = prop_match
    # query_parms_str = '&'.join(query_parms)
    # if query_parms_str:
    #     query_parms_str = f'?{query_parms_str}'

    return query_parms, client_filter_args


def make_query_str(query_parms):
    """
    Return the query parms as a string ready to be appended to a URI.

    If the list of query parms is empty, an empty string is returned.

    Parameters:

      query_parms (iterable of strings):
        List of query parameters, each being a string "name=value", where
        value is already URI-escaped if needed.

    Returns:

      str: query parms string
    """
    query_parms_str = '&'.join(query_parms)
    if query_parms_str:
        query_parms_str = f'?{query_parms_str}'
    return query_parms_str


def matches_filters(obj, filter_args):
    """
    Return a boolean indicating whether a resource object matches a set
    of filter arguments.

    This is used for client-side filtering.

    Depending on the properties specified in the filter arguments, this
    method retrieves the resource properties from the HMC.

    Parameters:

      obj (BaseResource):
        Resource object.

      filter_args (dict):
        Filter arguments. For details, see :ref:`Filtering`.
        `None` causes the resource to always match.

    Returns:

      bool: Boolean indicating whether the resource object matches the
        filter arguments.
    """
    if filter_args is not None:
        for prop_name in filter_args:
            prop_match = filter_args[prop_name]
            if prop_name == obj.manager.name_prop:
                case_insensitive = obj.manager.case_insensitive_names
            else:
                case_insensitive = False
            if not matches_prop(obj, prop_name, prop_match, case_insensitive):
                return False
    return True


def matches_prop(obj, prop_name, prop_match, case_insensitive):
    """
    Return a boolean indicating whether a resource object matches with
    a single property against a property match value.

    This is used for client-side filtering.

    Depending on the specified property, this method retrieves the resource
    properties from the HMC.

    Parameters:

      obj (BaseResource):
        Resource object.

      prop_match:
        Property match value that is used to match the actual value of
        the specified property against, as follows:

        - If the match value is a list or tuple, this method is invoked
          recursively to find whether one or more match values inthe list
          match.

        - Else if the property is of string type, its value is matched by
          interpreting the match value as a regular expression.

        - Else the property value is matched by exact value comparison
          with the match value.

      case_insensitive (bool):
        Controls whether the values of string typed properties are matched
        case insensitively.

    Returns:

      bool: Boolean indicating whether the resource object matches w.r.t.
        the specified property and the match value.
    """
    if isinstance(prop_match, (list, tuple)):
        # List items are logically ORed, so one matching item suffices.
        for pm in prop_match:
            if matches_prop(obj, prop_name, pm, case_insensitive):
                return True
    else:
        # Some lists of resources do not have all properties, for example
        # Hipersocket adapters do not have a "card-location" property.
        # If a filter property does not exist on a resource, the resource
        # does not match.
        try:
            prop_value = obj.get_property(prop_name)
        except KeyError:
            return False
        if isinstance(prop_value, str):
            # HMC resource property is Enum String or (non-enum) String,
            # and is both matched by regexp matching. Ideally, regexp
            # matching should only be done for non-enum strings, but
            # distinguishing them is not possible given that the client
            # has no knowledge about the properties.

            # The regexp matching implemented in the HMC requires begin and
            # end of the string value to match, even if the '^' for begin
            # and '$' for end are not specified in the pattern. The code
            # here is consistent with that: We add end matching to the
            # pattern, and begin matching is done by re.match()
            # automatically.
            re_match = prop_match + '$'
            re_flags = re.IGNORECASE if case_insensitive else 0
            m = re.match(re_match, prop_value, flags=re_flags)
            if m:
                return True
        else:
            if prop_value == prop_match:
                return True
    return False


def datetime_from_isoformat(dt_str):
    """
    Return a datetime object representing the date time string in ISO8601
    format.

    This function is used to parse timestamp strings from the externalized
    HMC definition.

    The date time strings are parsed using dateutil.parse.isoparse(), which
    supports the ISO8601 formats. The separator between the date portion and
    the time portion can be ' ' or 'T', and the optional timezone portion can
    be specified as 'shhmm' or 'shh:mm'.

    If the date time string specifies a timezone, the returned datetime object
    is timezone-aware. Otherwise, it is timezone-naive.
    """
    dt = parser.isoparse(dt_str)
    return dt


def datetime_to_isoformat(dt):
    """
    Return a date time string in ISO8601 format representing the datetime
    object.

    This function is used to create timestamp strings for the externalized
    HMC definition.

    The generated date time string has this format:

        YYYY-MM-DD HH:MM:SS[.ssssss][shh:mm]

    Where:
      * .ssssss - is an optional part specifying microseconds. It is not created
        when the datetime microsecond value is 0.
      * shh:mm - is an optional part specifying the timezone offset with sign,
        hours hh and minutes mm. It is not created when the datetime is
        timezone-naive.
    """
    dt_str = dt.isoformat(sep=' ')
    return dt_str


def get_features(session, base_uri, name):
    """
    Helper method that GETS the given uri via the given session, appending
    '/operations/list-features' to the uri (and the query parameter if needed).

    404/Not Found is caught and turned into an empty list result.
    """
    try:
        uri = f'{base_uri}/operations/list-features'
        if name is not None:
            uri = f'{uri}?name={name}'
        return session.get(uri)
    except HTTPError as e:
        # API features are introduced with WS API version 4.10.
        # Older HMCs will respond with 404/Not Found, which we simply
        # turn into "no features available at all".
        if e.http_status == 404:
            return []
        raise e


def warn_deprecated_parameter(cls, method, name, value, default):
    """
    Issue a DeprecationWarning for a zhmcclient method parameter.

    The method must use the @logged_api_call decorator.

    Parameters:

      cls (class): The class object that defines the method.

      method (method): The method object that has the parameter.

      name (str): The name of the parameter.

      value (object): The value of the parameter.

      default (object): The default value of the parameter.
    """
    if value != default:
        warnings.warn(
            f"Use of the '{name}' parameter of "
            f"zhmcclient.{cls.__name__}.{method.__name__}() has no "
            "function anymore and is deprecated",
            DeprecationWarning, stacklevel=5)
        # Note on stacklevel=5:
        # * base value 1 (get out of warnings.warn)
        # * +1 to get out of this function
        # * +1 to get out of method
        # * +2 to get over the @logged_api_call decorator of method


def stomp_uses_frames(stomp_version):
    """
    Returns whether stomp-py uses Frame objects for the event listener methods.

    Parameters:

      stomp_version: The __version__ attribute of the stomp module.
    """
    # stomp-py introduced the use of Frame objects in version 7.0.0, but since
    # it changed its __version__ attribute from tuple to string in version
    # 8.1.1, it would be fairly complex to check for the use of Frame objects
    # based upon its version. Instead, we utilize the fact that we have either
    # versions <5.0 or >8.1.1 and thus can test for the __version__ type.
    return isinstance(stomp_version, str)


def _add_if_set(kwargs, key, value):
    if value is not None:
        kwargs[key] = value


def get_stomp_rt_kwargs(rt_config):
    """
    Get the additional kwargs for StompConnection() from the
    stomp retry/timeout config.

    Parameters:

        rt_config (StompRetryTimeoutConfig): stomp retry/timeout config

    Returns:

        dict: Additional kwargs for StompConnection from the config.
    """
    rt_kwargs = dict()
    if rt_config:
        if rt_config.connect_timeout == 0:
            rt_kwargs['timeout'] = None  # No timeout
        elif rt_config.connect_timeout is not None:
            rt_kwargs['timeout'] = rt_config.connect_timeout
        _add_if_set(rt_kwargs, 'reconnect_attempts_max',
                    rt_config.connect_retries)
        _add_if_set(rt_kwargs, 'reconnect_sleep_initial',
                    rt_config.reconnect_sleep_initial)
        _add_if_set(rt_kwargs, 'reconnect_sleep_increase',
                    rt_config.reconnect_sleep_increase)
        _add_if_set(rt_kwargs, 'reconnect_sleep_max',
                    rt_config.reconnect_sleep_max)
        _add_if_set(rt_kwargs, 'reconnect_sleep_jitter',
                    rt_config.reconnect_sleep_jitter)
        if rt_config.keepalive is False:
            rt_kwargs['keepalive'] = None  # Disabled
        elif rt_config.keepalive is not None:
            rt_kwargs['keepalive'] = rt_config.keepalive
        if rt_config.heartbeat_send_cycle is not None or \
                rt_config.heartbeat_receive_cycle is not None:
            send_cycle = int(rt_config.heartbeat_send_cycle * 1000) or 0
            rcv_cycle = int(rt_config.heartbeat_receive_cycle * 1000) or 0
            rt_kwargs['heartbeats'] = (send_cycle, rcv_cycle)
        if rt_config.heartbeat_receive_check is not None:
            rcv_scale = 1.0 + rt_config.heartbeat_receive_check
            rt_kwargs['heart_beat_receive_scale'] = rcv_scale
    return rt_kwargs


def get_headers_message(frame_args):
    """
    Transform STOMP event method parameters to a tuple(headers, message),
    dependent on the stomp-py version that is used.

    Parameters:

      frame_args: The STOMP frame, represented depending on the stomp-py
        package version as follows:

          * if stomp-py uses Frames, a single stomp.Frame object:

            frame (stomp.Frame): Object with STOMP message headers and
              message body.

          * else, a tuple(headers, message):

            headers (dict): STOMP message headers.
              The headers are described in the `headers` tuple item
              returned by the
              :meth:`~zhmcclient.NotificationReceiver.notifications`
              method.

            message (string): STOMP message body as a string, which
              contains a serialized JSON object.
              The JSON object is described in the `message` tuple item
              returned by the
              :meth:`~zhmcclient.NotificationReceiver.notifications`
              method.
    """
    if len(frame_args) == 1:
        frame = frame_args[0]
        headers = frame.headers
        message = frame.body
    else:
        headers, message = frame_args
    return headers, message



#!/usr/bin/env python
# Copyright 2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that shows the auto-updating of a partition manager on a CPC in DPM
mode.

The shorter durations of list() indicate that the locally maintained list was
used; the longer durations indicate that a list operation was performed on the
HMC.
"""

import sys
import uuid
import logging
import requests.packages.urllib3
from datetime import datetime, timedelta

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

# Flag to control logging of the HTTP and notification interactions with
# the HMC to stderr.
ENABLE_LOGGING = False

if ENABLE_LOGGING:
    logging.basicConfig(level=logging.DEBUG, stream=sys.stderr)


def delta_ms(start_dt, end_dt):
    """Return float that is the time delta in milliseconds"""
    return (end_dt - start_dt) / timedelta(microseconds=1) / 1000


print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

cleanup_partition = None
try:
    client = zhmcclient.Client(session)

    print("Finding a CPC in DPM mode ...")
    cpcs = client.cpcs.list(filter_args={'dpm-enabled': True})
    if not cpcs:
        print(f"Error: HMC at {host} does not manage any CPCs in DPM mode")
        sys.exit(1)
    cpc = cpcs[0]
    print(f"Using CPC {cpc.name}")

    part_mgr = cpc.partitions

    print("Listing partitions using list() (not auto-enabled) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

    print("Enabling auto-updating for partition manager")
    try:
        part_mgr.enable_auto_update()
    except zhmcclient.Error as exc:
        print(f"Error: Cannot auto-enable partition manager for CPC "
              f"{cpc.name}: {exc.__class__.__name__}: {exc}")
        sys.exit(1)

    print("Listing partitions using list() "
          "(auto-enabled - uses from local list) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

    part_name = f"zhmc_test_{uuid.uuid4()}"
    print(f"Creating partition {part_name} ...")
    part_props = {
        'name': part_name,
        'description': 'Original partition description.',
        'cp-processors': 2,
        'initial-memory': 1024,
        'maximum-memory': 2048,
        'processor-mode': 'shared'
    }
    try:
        part = cpc.partitions.create(properties=part_props)
        cleanup_partition = part
    except zhmcclient.Error as exc:
        print(f"Error: Cannot create partition {part_name} on CPC {cpc.name}: "
              f"{exc.__class__.__name__}: {exc}")
        sys.exit(1)
    print(f"Partition uri: {part.uri}")

    print("Listing partitions using list() "
          "(auto-enabled - added partition causes pull from HMC) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

    print("Listing partitions using list() "
          "(auto-enabled - uses from local list) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

    print(f"Deleting partition {part.name} (uri: {part.uri}) ...")
    try:
        part.delete()
        cleanup_partition = None
    except zhmcclient.Error as exc:
        print(f"Error: Cannot delete partition {part.name} on CPC {cpc.name} "
              "for clean up - Please delete it manually: "
              f"{exc.__class__.__name__}: {exc}")
        sys.exit(1)

    print("Listing partitions using list() "
          "(auto-enabled - uses from local list) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

    print("Disabling auto-updating for partition manager")
    part_mgr.disable_auto_update()

    print("Listing partitions using list() (not auto-enabled) ...")
    start_dt = datetime.now()
    part_list = part_mgr.list()
    end_dt = datetime.now()
    print(f"Result of list(): returned {len(part_list)} partitions in "
          f"{delta_ms(start_dt, end_dt)} ms")

finally:
    if cleanup_partition:
        print(f"Cleanup: Deleting partition {cleanup_partition.name} ...")
        try:
            cleanup_partition.delete()
        except zhmcclient.Error as exc:
            print(f"Error: Cannot delete partition {part.name} on CPC "
                  f"{cpc.name} for clean up - "
                  f"Please delete it manually: {exc.__class__.__name__}: {exc}")
    print("Logging off ...")
    session.logoff()



#!/usr/bin/env python
# Copyright 2016,2022 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Example that lists the CPCs managed by the HMC.
"""

import sys
import requests.packages.urllib3

import zhmcclient
from zhmcclient.testutils import hmc_definitions

requests.packages.urllib3.disable_warnings()

# Get HMC info from HMC inventory and vault files
hmc_def = hmc_definitions()[0]
nickname = hmc_def.nickname
host = hmc_def.host
userid = hmc_def.userid
password = hmc_def.password
verify_cert = hmc_def.verify_cert

print(__doc__)

print(f"Using HMC {nickname} at {host} with userid {userid} ...")

print("Creating a session with the HMC ...")
try:
    session = zhmcclient.Session(
        host, userid, password, verify_cert=verify_cert)
except zhmcclient.Error as exc:
    print(f"Error: Cannot establish session with HMC {host}: "
          f"{exc.__class__.__name__}: {exc}")
    sys.exit(1)

try:
    client = zhmcclient.Client(session)
    format_str = "{:<8}  {:<6}  {:<7}  {:<16}"

    cpcs = client.cpcs.list()
    print('')
    print('CPCs managed by this HMC:')
    print(format_str.format('CPC', 'SE', 'Mode', 'Status'))
    for cpc in sorted(cpcs, key=lambda x: x.name):
        mode = 'DPM' if cpc.prop('dpm-enabled') else 'classic'
        print(format_str.format(
            cpc.name, cpc.prop('se-version'), mode, cpc.prop('status')))
    print('')

finally:
    print("Logging off ...")
    session.logoff()



# Copyright 2016,2023 IBM Corp. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A :term:`Certificate` represents an X.509 certificate.

Certificates are top level resources, but at this point, only certificates of
type "secure-boot" are supported. Such certificates are always associated to a
specific :term:`CPC`. They can only be used if the
:ref:`API feature <API features>` "secure-boot-with-certificates" is available
on both, :term:`HMC` and :term:`CPC`.

They can be assigned to one or more :term:`LPAR`, :term:`Partition`, or
:term:`Image Activation Profile` to be used during the load/start of the
corresponding entities when doing a "secure boot" load for an LPAR, respectively
a Partition start.
"""


import copy

from ._logging import logged_api_call
from ._manager import BaseManager
from ._resource import BaseResource
from ._utils import RC_CERTIFICATE

__all__ = ['CertificateManager', 'Certificate']


class CertificateManager(BaseManager):
    """
    Manager providing access to the :term:`Certificates <Certificate>` of an
    :term:`HMC`.

    Derived from :class:`~zhmcclient.BaseManager`; see there for common methods
    and attributes.

    Objects of this class are not directly created by the user; they are
    accessible via the following instance variable:

    * :attr:`~zhmcclient.Console.certificates` of a :class:`~zhmcclient.Console`
      object.

    HMC/SE version requirements:

    * :ref:`API feature <API features>` "secure-boot-with-certificates"
    """

    def __init__(self, console):
        # This function should not go into the docs.
        # Parameters:
        #   console (:class:`~zhmcclient.Console`):
        #     HMC defining the scope for this manager.

        # Resource properties that are supported as filter query parameters.
        # If the support for a resource property changes within the set of HMC
        # versions that support this type of resource, this list must be set up
        # for the version of the HMC this session is connected to.
        query_props = [
            'name', 'parent-name', 'type'
        ]

        super().__init__(
            resource_class=Certificate,
            class_name=RC_CERTIFICATE,
            session=console.manager.session,
            parent=console,
            base_uri='/api/certificates',
            oid_prop='object-id',
            uri_prop='object-uri',
            name_prop='name',
            query_props=query_props)

    @property
    def console(self):
        """
        :class:`~zhmcclient.Console`: The Console object representing the HMC.
        """
        return self._parent

    @logged_api_call
    # pylint: disable=arguments-differ
    def list(self, full_properties=False, filter_args=None,
             additional_properties=None):
        """
        List the certificates defined in the HMC.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to any Certificate to be included in the
          result.

        Parameters:

          full_properties (bool):
            Controls whether the full set of resource properties should be
            retrieved, vs. only the short set as returned by the list
            operation.

          filter_args (dict):
            Filter arguments that narrow the list of returned resources to
            those that match the specified filter arguments. For details, see
            :ref:`Filtering`.

            `None` causes no filtering to happen, i.e. all resources are
            returned.

          additional_properties (list of string):
            List of property names that are to be returned in addition to the
            default properties.

        Returns:

          : A list of :class:`~zhmcclient.Certificate` objects.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result_prop = 'certificates'
        list_uri = '/api/certificates'
        return self._list_with_operation(
            list_uri, result_prop, full_properties, filter_args,
            additional_properties)

    @logged_api_call
    def import_certificate(self, cpc, properties):
        """
        Imports a certificate to a CPC.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to the given CPC.
        * Task permission for the "Import Secure Boot Certificates" task.

        Parameters:

          cpc (:class:`~zhmcclient.Cpc`): the CPC the certificate will be
          imported to.

          properties (dict): Initial property values.
            Allowable properties are defined in section 'Request body contents'
            in section 'Import CPC Certificate' in the :term:`HMC API` book.

        Returns:

          :class:`~zhmcclient.Certificate`:
            The resource object for the new Certificate.
            The object will have its 'object-uri' property set as returned by
            the HMC, and will also have the input properties set.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        result = self.session.post(
            cpc.uri + '/operations/import-certificate', body=properties)
        # There should not be overlaps, but just in case there are, the
        # returned props should overwrite the input props:
        props = copy.deepcopy(properties)
        props.update(result)
        name = props.get(self._name_prop, None)
        # import result uses a different key to identify the uri
        uri = props['certificate-uri']
        cert = Certificate(self, uri, name, props)
        self._name_uri_cache.update(name, uri)
        return cert


class Certificate(BaseResource):
    """
    Representation of a :term:`Certificate`.

    Derived from :class:`~zhmcclient.BaseResource`; see there for common
    methods and attributes.

    Objects of this class are not directly created by the user; they are
    returned from creation or list functions on their manager object
    (in this case, :class:`~zhmcclient.CertificateManager`).

    HMC/SE version requirements:

    * :ref:`API feature <API features>` "secure-boot-with-certificates"
    """

    def __init__(self, manager, uri, name=None, properties=None):
        # This function should not go into the docs.
        #   manager (:class:`~zhmcclient.CertificateManager`):
        #     Manager object for this resource object.
        #   uri (string):
        #     Canonical URI path of the resource.
        #   name (string):
        #     Name of the resource.
        #   properties (dict):
        #     Properties to be set for this resource object. May be `None` or
        #     empty.
        assert isinstance(manager, CertificateManager), (
            f"Certificate init: Expected manager type {CertificateManager}, "
            f"got {type(manager)}")
        super().__init__(manager, uri, name, properties)
        self._cpc = None

    @property
    def cpc(self):
        """
        :class:`~zhmcclient.Cpc`: The :term:`CPC` to which this storage group
        is associated.

        The returned :class:`~zhmcclient.Cpc` has only a minimal set of
        properties populated.
        """
        # We do here some lazy loading.
        if not self._cpc:
            cpc_uri = self.get_property('parent')
            cpc_mgr = self.manager.console.manager.client.cpcs
            self._cpc = cpc_mgr.resource_object(cpc_uri)
        return self._cpc

    @logged_api_call
    def delete(self):
        """
        Delete this Certificate.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this Certificate.
        * Task permission to the "Import Secure Boot Certificates" task.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.delete(self.uri, resource=self)
        self.manager._name_uri_cache.delete(
            self.get_properties_local(self.manager._name_prop, None))
        self.cease_existence_local()

    @logged_api_call
    def update_properties(self, properties):
        """
        Update writeable properties of this Certificate.

        This method serializes with other methods that access or change
        properties on the same Python object.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this Certificate.
        * Task permission to the "Certificate Details" task.

        Parameters:

          properties (dict): New values for the properties to be updated.
            Properties not to be updated are omitted.
            Allowable properties are the properties with qualifier (w) in
            section 'Data model' in section 'Certificate object' in the
            :term:`HMC API` book.

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        self.manager.session.post(self.uri, resource=self, body=properties)
        is_rename = self.manager._name_prop in properties
        if is_rename:
            # Delete the old name from the cache
            self.manager._name_uri_cache.delete(self.name)
        self.update_properties_local(copy.deepcopy(properties))
        if is_rename:
            # Add the new name to the cache
            self.manager._name_uri_cache.update(self.name, self.uri)

    @logged_api_call
    def get_encoded(self):
        """
        Gets the Base64 encoded string and the file format of this certificate.

        HMC/SE version requirements:

        * :ref:`API feature <API features>` "secure-boot-with-certificates"

        Authorization requirements:

        * Object-access permission to this Certificate.

        Returns:

            dict: The encoded certificate in conjunction with its format:

            * key "certificate": The Base64 encoded string of the certificate.
            * key "format": The format of the certificate ("der" or "pem").

        Raises:

          :exc:`~zhmcclient.HTTPError`
          :exc:`~zhmcclient.ParseError`
          :exc:`~zhmcclient.AuthError`
          :exc:`~zhmcclient.ConnectionError`
        """
        # pylint: disable=protected-access
        return self.manager.session.get(
            f'{self.uri}/operations/get-encoded', resource=self)

    def dump(self):
        """
        Dump this Certificate resource with its properties as a resource
        definition.

        The returned resource definition has the following format::

            {
                # Resource properties:
                "properties": {...},
            }

        Returns:

          dict: Resource definition of this resource.
        """

        # Dump the resource properties
        resource_dict = super().dump()

        return resource_dict
